# =====================================
# generator=datazen
# version=3.1.3
# hash=50538057667f048e4e1c2e1232e77e1b
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "2.4.0"



"""
A module implementing a data model for ARM CMSIS-SVD 'peripheral' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable, List, Optional, Tuple
from xml.etree import ElementTree

# internal
from ifgen.svd.model.address_block import AddressBlock
from ifgen.svd.model.cluster import ClusterMap, get_clusters
from ifgen.svd.model.derived import DerivedMixin
from ifgen.svd.model.device import ARRAY_PROPERTIES, REGISTER_PROPERTIES
from ifgen.svd.model.interrupt import Interrupt
from ifgen.svd.model.register import RegisterMap, get_registers
from ifgen.svd.string import StringKeyVal

RegisterData = Tuple[ClusterMap, RegisterMap]


@dataclass
class Peripheral(DerivedMixin):
    """A container for peripheral information."""

    derived_from: Optional["Peripheral"]
    interrupts: List[Interrupt]
    address_blocks: List[AddressBlock]
    registers: List[RegisterData]

    def handle_registers(self, registers: ElementTree.Element) -> None:
        """Handle the 'registers' element."""

        self.registers.append(
            (get_clusters(registers), get_registers(registers))
        )

    def handle_address_block(self, address_block: ElementTree.Element) -> None:
        """Handle an 'address_block' element."""
        self.address_blocks.append(AddressBlock.create(address_block))

    def handle_interrupt(self, interrupt: ElementTree.Element) -> None:
        """Handle an 'interrupt' element."""
        self.interrupts.append(Interrupt.create(interrupt))

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return (
            ARRAY_PROPERTIES
            + [
                StringKeyVal("name", True),
                StringKeyVal("version", False),
                StringKeyVal("description", False),
                StringKeyVal("alternatePeripheral", False),
                StringKeyVal("groupName", False),
                StringKeyVal("prependToName", False),
                StringKeyVal("appendToName", False),
                StringKeyVal("headerStructName", False),
                StringKeyVal("disableCondition", False),
                StringKeyVal("baseAddress", True),
            ]
            + REGISTER_PROPERTIES
        )



"""
A module implementing interfaces for struct-file generation.
"""

# built-in
from typing import Any, Dict, Iterable, Union

# third-party
from vcorelib.io.file_writer import (
    CommentStyle,
    IndentedFileWriter,
    LineWithComment,
)

# internal
from ifgen import PKG_NAME
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods import struct_methods
from ifgen.struct.methods.fields import bit_fields
from ifgen.struct.source import create_struct_source
from ifgen.struct.stream import struct_stream_methods
from ifgen.struct.test import create_struct_test

__all__ = ["create_struct", "create_struct_test", "create_struct_source"]
FieldConfig = Dict[str, Union[int, str]]


def struct_line(
    name: str, value: FieldConfig, volatile: bool
) -> LineWithComment:
    """Build a string for a struct-field line."""

    return ("volatile " if volatile else "") + (  # type: ignore
        f"{value['type']} {name};"
    ), value.get("description")


def header_for_type(name: str, task: GenerateTask) -> str:
    """Determine the header file to import for a given type."""

    candidate = task.custom_include(name)
    if candidate:
        return f'"{candidate}"'

    return ""


def struct_includes(task: GenerateTask) -> Iterable[str]:
    """Determine headers that need to be included for a given struct."""

    result = {
        header_for_type(config["type"], task)
        for config in task.instance["fields"]
    }

    result.add(f'"../{PKG_NAME}/common.h"')

    return result


def struct_fields(task: GenerateTask, writer: IndentedFileWriter) -> None:
    """Generate struct fields."""

    writer.c_comment("Fields.")

    with writer.trailing_comment_lines(style=CommentStyle.C_DOXYGEN) as lines:
        # Fields.
        for field in task.instance["fields"]:
            lines.append(struct_line(field["name"], field, field["volatile"]))

        lines.append(("", None))


def struct_instance(
    task: GenerateTask, writer: IndentedFileWriter, instance: dict[str, Any]
) -> None:
    """Generate struct instances."""

    writer.empty()

    if instance.get("description"):
        with writer.javadoc():
            writer.write(instance["description"])

    type_base = f"{task.name} *"
    writer.write(
        (
            "static "
            + ("volatile " if instance["volatile"] else "")
            + f"{type_base}const "
            f"{instance['name']} = "
            f"reinterpret_cast<{type_base}>({instance['address']});"
        )
    )


def create_struct(task: GenerateTask) -> None:
    """Create a header file based on a struct definition."""

    with task.boilerplate(
        includes=struct_includes(task), json=task.instance.get("json", False)
    ) as writer:
        attributes = ["gnu::packed"]
        writer.write(f"struct [[{', '.join(attributes)}]] {task.name}")
        with writer.scope(suffix=";"):
            writer.c_comment("Constant attributes.")
            with writer.trailing_comment_lines(
                style=CommentStyle.C_DOXYGEN
            ) as lines:
                lines.append(
                    (
                        "static constexpr "
                        f"{task.env.config.data['struct_id_underlying']} "
                        f"id = {task.protocol().id};",
                        f"{task.name}'s identifier.",
                    )
                )
                lines.append(
                    (
                        f"static constexpr std::size_t size = "
                        f"{task.env.types.size(task.name)};",
                        f"{task.name}'s size in bytes.",
                    )
                )

            writer.empty()
            struct_fields(task, writer)

            # Methods.
            writer.c_comment("Methods.")
            struct_methods(task, writer, True)
            bit_fields(task, writer, True)

        # Add size assertion.
        writer.empty()
        writer.write(
            f"static_assert(sizeof({task.name}) == {task.name}::size);"
        )

        if task.instance["stream"]:
            writer.empty()
            struct_stream_methods(task, writer, True)

        for instance in task.instance.get("instances", []):
            struct_instance(task, writer, instance)



"""
A module implementing an interface for generating bit-field methods for
structs.
"""

# built-in
from contextlib import ExitStack
from typing import Any

# third-party
from vcorelib.io.file_writer import IndentedFileWriter

# internal
from ifgen.generation.interface import GenerateTask
from ifgen.struct.methods.bit import bit_field_toggle_method

STANDARD_INTS = [
    ("uint8_t", 8),
    ("uint16_t", 16),
    ("uint32_t", 32),
    ("uint64_t", 64),
]


def bit_field_underlying(field: dict[str, Any]) -> str:
    """Get the underlying type for a bit field."""

    kind = field.get("type")

    # Automatically determine a sane primitive-integer type to use if one isn't
    # specified.
    if kind is None:
        width = field["width"]

        if width == 1:
            kind = "bool"
        else:
            for candidate, bit_width in STANDARD_INTS:
                if field["width"] <= bit_width:
                    kind = candidate
                    break

    assert kind is not None, kind
    return kind


def bit_field_get_method(
    task: GenerateTask,
    parent: dict[str, Any],
    field: dict[str, Any],
    writer: IndentedFileWriter,
    header: bool,
    kind: str,
    method_slug: str,
) -> None:
    """Generate a 'get' method for a bit-field."""

    if not header:
        return

    is_flag = field["width"] == 1

    method = task.cpp_namespace(f"get_{method_slug}()", header=header)
    writer.empty()

    with writer.javadoc():
        writer.write(
            (
                f"Get {parent['name']}'s {field['name']} "
                f"{'field' if field['width'] > 1 else 'bit'}."
            )
        )

    line = f"{kind} " + method

    with ExitStack() as stack:
        if is_flag:
            writer.write(line)
            stack.enter_context(writer.scope())
            stmt = f"{parent['name']} & (1 << {field['index']})"
        else:
            writer.write(line)
            stack.enter_context(writer.scope())
            mask = bit_mask_literal(field["width"])
            stmt = f"({parent['name']} >> {field['index']}u) & {mask}"

        if task.env.is_enum(kind):
            stmt = f"{kind}({stmt})"

        writer.write(f"return {stmt};")


def bit_mask_literal(width: int) -> str:
    """Get a bit-mask literal."""
    return "0b" + ("1" * width) + "u"


def bit_field_set_method(
    task: GenerateTask,
    parent: dict[str, Any],
    field: dict[str, Any],
    writer: IndentedFileWriter,
    header: bool,
    kind: str,
    method_slug: str,
) -> None:
    """Generate a 'set' method for a bit-field."""

    # Generate a toggle method for bit fields.
    if field["width"] == 1:
        bit_field_toggle_method(
            task, parent["name"], field, writer, header, method_slug
        )
    else:
        if not header:
            return

        method = task.cpp_namespace(
            f"set_{method_slug}({kind} value)", header=header
        )
        writer.empty()

        if header:
            with writer.javadoc():
                writer.write(f"Set {parent['name']}'s {field['name']} field.")

        writer.write("inline void " + method)
        with writer.scope():
            writer.write(f"{parent['type']} curr = {parent['name']};")

            mask = bit_mask_literal(field["width"])

            with writer.padding():
                writer.write(f"curr &= ~({mask} << {field['index']}u);")

                val_str = "value"
                if task.env.is_enum(kind):
                    val_str = f"std::to_underlying({val_str})"

                writer.write(
                    f"curr |= ({val_str} & {mask}) << {field['index']}u;"
                )

            writer.write(f"{parent['name']} = curr;")


def bit_field(
    task: GenerateTask,
    parent: dict[str, Any],
    field: dict[str, Any],
    writer: IndentedFileWriter,
    header: bool,
) -> None:
    """Generate for an individual bit-field."""

    kind = bit_field_underlying(field)

    type_size = task.env.size(parent["type"]) * 8

    index = field["index"]
    width = field["width"]

    # Validate field parameters.
    assert index + width < type_size, (index, width, type_size, field)
    assert field["read"] or field["write"], field

    name = parent["name"]
    method_slug = f"{name}_{field['name']}"

    # Generate a 'get' method.
    if field["read"]:
        bit_field_get_method(
            task, parent, field, writer, header, kind, method_slug
        )

    # Generate a 'set' method.
    if field["write"]:
        bit_field_set_method(
            task, parent, field, writer, header, kind, method_slug
        )


def bit_fields(
    task: GenerateTask, writer: IndentedFileWriter, header: bool
) -> None:
    """Generate bit-field lines."""

    for field in task.instance["fields"]:
        for bfield in field.get("fields", []):
            bit_field(task, field, bfield, writer, header)



"""
A module implementing a data model for ARM CMSIS-SVD 'field' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable, Optional
from xml.etree import ElementTree

# internal
from ifgen.svd.model.derived import DerivedMixin, derived_from_stack
from ifgen.svd.model.device import ARRAY_PROPERTIES
from ifgen.svd.model.enum import EnumeratedValues, get_enum
from ifgen.svd.string import StringKeyVal


@dataclass
class Field(DerivedMixin):
    """A container for field information."""

    derived_from: Optional["Field"]
    enum: Optional[EnumeratedValues]

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        # Not currently handling nested registers or clusters.

        return ARRAY_PROPERTIES + [
            StringKeyVal("name", True),
            StringKeyVal("description", False),
            # Mutually exclusive options:
            # 1 (bitRangeLsbMsbStyle)
            StringKeyVal("bitOffset", False),
            StringKeyVal("bitWidth", False),
            # 2 (bitRangeOffsetWidthStyle)
            StringKeyVal("lsb", False),
            StringKeyVal("msb", False),
            # 3 (bitRangePattern)
            StringKeyVal("bitRange", False),
            # (end mutually exclusive options)
            # is enum
            StringKeyVal("access", False),
            StringKeyVal("modifiedWriteValues", False),
            # is enum
            StringKeyVal("readAction", False),
        ]


FieldMap = dict[str, Field]


def get_fields(registers: ElementTree.Element) -> FieldMap:
    """Get register fields."""

    result: FieldMap = {}
    for field in derived_from_stack(registers.iterfind("field")):
        derived_field = None
        derived = field.attrib.get("derivedFrom")
        if derived is not None:
            derived_field = result[derived]  # pragma: nocover

        # Handle writeConstraint at some point?

        # Handle enumerated values.
        enum = None
        enum_elem = field.find("enumeratedValues")
        if enum_elem is not None:
            enum = get_enum(enum_elem)

        inst = Field.create(field, derived_field, enum)
        result[inst.name] = inst

    return result



"""
A module implementing a data model for ARM CMSIS-SVD 'cluster' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable, Optional
from xml.etree import ElementTree

# internal
from ifgen.svd.model.derived import DerivedMixin, derived_from_stack
from ifgen.svd.model.device import ARRAY_PROPERTIES, REGISTER_PROPERTIES
from ifgen.svd.string import StringKeyVal


@dataclass
class Cluster(DerivedMixin):
    """A container for cluster information."""

    derived_from: Optional["Cluster"]

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        # Not currently handling nested registers or clusters.

        return (
            ARRAY_PROPERTIES
            + [
                StringKeyVal("name", True),
                StringKeyVal("description", False),
                StringKeyVal("alternateCluster", False),
                StringKeyVal("headerStructName", False),
                StringKeyVal("addressOffset", True),
            ]
            + REGISTER_PROPERTIES
        )


ClusterMap = dict[str, Cluster]


def get_clusters(registers: ElementTree.Element) -> ClusterMap:
    """Get register clusters."""

    result: ClusterMap = {}
    for cluster in derived_from_stack(registers.iterfind("cluster")):
        derived_cluster = None
        derived = cluster.attrib.get("derivedFrom")
        if derived is not None:
            derived_cluster = result[derived]  # pragma: nocover

        inst = Cluster.create(cluster, derived_cluster)
        result[inst.name] = inst

    return result



"""
A module implementing an SVD-processing task interface.
"""

# built-in
from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import Callable, Dict
from xml.etree import ElementTree

# third-party
from vcorelib.logging import LoggerType

# internal
from ifgen.svd.model import SvdModel

TagProcessor = Callable[
    [ElementTree.Element, "SvdProcessingTask", LoggerType], None
]
TagProcessorMap = Dict[str, TagProcessor]
TAG_PROCESSORS: TagProcessorMap = {}


@dataclass
class SvdProcessingTask:
    """A container for SVD-processing state."""

    model: SvdModel

    def process(self, elem: ElementTree.Element) -> None:
        """Process a single element."""
        TAG_PROCESSORS[elem.tag](elem, self, getLogger(elem.tag))

    @staticmethod
    def svd(path: Path) -> "SvdProcessingTask":
        """Process a single SVD file."""

        task = SvdProcessingTask(SvdModel({}))
        task.process(ElementTree.parse(path).getroot())
        return task



"""
A module implementing a data model for ARM CMSIS-SVD 'register' data.
"""

# built-in
from dataclasses import dataclass
from typing import Iterable, Optional
from xml.etree import ElementTree

# internal
from ifgen.svd.model.derived import DerivedMixin, derived_from_stack
from ifgen.svd.model.device import ARRAY_PROPERTIES, REGISTER_PROPERTIES
from ifgen.svd.model.field import FieldMap, get_fields
from ifgen.svd.string import StringKeyVal


@dataclass
class Register(DerivedMixin):
    """A container for register information."""

    derived_from: Optional["Register"]
    fields: Optional[FieldMap]

    @classmethod
    def string_keys(cls) -> Iterable[StringKeyVal]:
        """Get string keys for this instance type."""

        return (
            ARRAY_PROPERTIES
            + [
                StringKeyVal("name", True),
                StringKeyVal("displayName", False),
                StringKeyVal("description", False),
                StringKeyVal("alternateGroup", False),
                StringKeyVal("alternateRegister", False),
                StringKeyVal("addressOffset", True),
            ]
            + REGISTER_PROPERTIES
            + [
                # is enum
                StringKeyVal("dataType", False),
                # is enum
                StringKeyVal("modifiedWriteValues", False),
                # is enum
                StringKeyVal("readAction", False),
            ]
        )


RegisterMap = dict[str, Register]


def get_registers(registers: ElementTree.Element) -> RegisterMap:
    """Get register elements."""

    result: RegisterMap = {}
    for register in derived_from_stack(registers.iterfind("register")):
        derived_register = None
        derived = register.attrib.get("derivedFrom")
        if derived is not None:
            derived_register = result[derived]  # pragma: nocover

        # Handle writeConstraint at some point?

        # Load fields.
        fields = None
        fields_elem = register.find("fields")
        if fields_elem is not None:
            fields = get_fields(fields_elem)

        inst = Register.create(register, derived_register, fields)
        result[inst.name] = inst

    return result
